from typing import Optional

from openai.types.chat import ChatCompletionAssistantMessageParam


class NexusflowAIChatCompletionAssistantMessageParam(
    ChatCompletionAssistantMessageParam
):
    nexusflowai_extras: Optional[str] = None
    """Extra items passed in for Nexusflow client, formatted as a json string."""



from nexusflowai.resources.completions import (
    Completions,
    AsyncCompletions,
    CompletionsWithRawResponse,
    AsyncCompletionsWithRawResponse,
)
from nexusflowai.resources.chat import (
    Chat,
    AsyncChat,
    ChatWithRawResponse,
    AsyncChatWithRawResponse,
)

from nexusflowai.resources.models import (
    Models,
    AsyncModels,
    ModelsWithRawResponse,
    AsyncModelsWithRawResponse,
)



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Union, Optional
from typing_extensions import Literal

import httpx

from openai._types import NOT_GIVEN, Body, Query, Headers, NotGiven
from openai._utils import required_args, maybe_transform
from openai._resource import SyncAPIResource, AsyncAPIResource
from openai._response import to_raw_response_wrapper, async_to_raw_response_wrapper
from openai._streaming import Stream, AsyncStream
from openai.types.chat import (
    ChatCompletionToolParam,
    ChatCompletionMessageParam,
    ChatCompletionToolChoiceOptionParam,
    completion_create_params,
)
from openai._base_client import make_request_options

from openai.types.chat.completion_create_params import (
    CompletionCreateParams as ChatCompletionCreateParams,
)
from openai.types.chat.chat_completion_message import ChatCompletionMessage

from nexusflowai.types import NexusflowAIChatCompletion, NexusflowAIChatCompletionChunk
from nexusflowai.validators import ChatCompletionsInputValidator, QwenAgentFlow
from nexusflowai.utils import ALLOWED_MODELS, transform_client_header
from nexusflowai.resources.utils import (
    maybe_print_hints,
    maybe_parse_into_pydantic,
    batch_call,
)


if TYPE_CHECKING:
    from nexusflowai.client import NexusflowAI, AsyncNexusflowAI

__all__ = ["Completions", "AsyncCompletions", "ChatCompletionsInputValidator"]


class Completions(SyncAPIResource):
    with_raw_response: CompletionsWithRawResponse

    def __init__(self, client: NexusflowAI) -> None:
        super().__init__(client)
        self.with_raw_response = CompletionsWithRawResponse(self)

    @required_args(["messages", "model"], ["messages", "model", "stream"])
    def create(
        self,
        *,
        messages: List[ChatCompletionMessageParam],
        model: Union[
            str,
            ALLOWED_MODELS,
        ],
        frequency_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        function_call: completion_create_params.FunctionCall | NotGiven = NOT_GIVEN,
        functions: List[completion_create_params.Function] | NotGiven = NOT_GIVEN,
        logit_bias: Optional[Dict[str, int]] | NotGiven = NOT_GIVEN,
        max_tokens: Optional[int] | NotGiven = NOT_GIVEN,
        n: Optional[int] | NotGiven = NOT_GIVEN,
        parallel_tool_calls: bool | NotGiven = NOT_GIVEN,
        presence_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        response_format: completion_create_params.ResponseFormat | NotGiven = NOT_GIVEN,
        seed: Optional[int] | NotGiven = NOT_GIVEN,
        stop: Union[Optional[str], List[str]] | NotGiven = NOT_GIVEN,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        temperature: Optional[float] | NotGiven = NOT_GIVEN,
        tool_choice: ChatCompletionToolChoiceOptionParam | NotGiven = NOT_GIVEN,
        tools: List[ChatCompletionToolParam] | NotGiven = NOT_GIVEN,
        top_p: Optional[float] | NotGiven = NOT_GIVEN,
        user: str | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> NexusflowAIChatCompletion | Stream[NexusflowAIChatCompletionChunk]:
        """
        Creates a model response for the given chat conversation.

        Args:
          messages: A list of messages comprising the conversation so far.
              [Example Python code](https://cookbook.openai.com/examples/how_to_format_inputs_to_chatgpt_models).

          model: ID of the model to use. See the
              [model endpoint compatibility](https://platform.openai.com/docs/models/model-endpoint-compatibility)
              table for details on which models work with the Chat API.

          frequency_penalty: Number between -2.0 and 2.0. Positive values penalize new tokens based on their
              existing frequency in the text so far, decreasing the model's likelihood to
              repeat the same line verbatim.

              [See more information about frequency and presence penalties.](https://platform.openai.com/docs/guides/gpt/parameter-details)

          function_call: Deprecated in favor of `tool_choice`.

              Controls which (if any) function is called by the model. `none` means the model
              will not call a function and instead generates a message. `auto` means the model
              can pick between generating a message or calling a function. Specifying a
              particular function via `{"name": "my_function"}` forces the model to call that
              function.

              `none` is the default when no functions are present. `auto`` is the default if
              functions are present.

          functions: Deprecated in favor of `tools`.

              A list of functions the model may generate JSON inputs for.

          logit_bias: Modify the likelihood of specified tokens appearing in the completion.

              Accepts a JSON object that maps tokens (specified by their token ID in the
              tokenizer) to an associated bias value from -100 to 100. Mathematically, the
              bias is added to the logits generated by the model prior to sampling. The exact
              effect will vary per model, but values between -1 and 1 should decrease or
              increase likelihood of selection; values like -100 or 100 should result in a ban
              or exclusive selection of the relevant token.

          max_tokens: The maximum number of [tokens](/tokenizer) to generate in the chat completion.

              The total length of input tokens and generated tokens is limited by the model's
              context length.
              [Example Python code](https://cookbook.openai.com/examples/how_to_count_tokens_with_tiktoken)
              for counting tokens.

          n: How many chat completion choices to generate for each input message.

          presence_penalty: Number between -2.0 and 2.0. Positive values penalize new tokens based on
              whether they appear in the text so far, increasing the model's likelihood to
              talk about new topics.

              [See more information about frequency and presence penalties.](https://platform.openai.com/docs/guides/gpt/parameter-details)

          response_format: An object specifying the format that the model must output.

              Setting to `{ "type": "json_object" }` enables JSON mode, which guarantees the
              message the model generates is valid JSON.

              **Important:** when using JSON mode, you **must** also instruct the model to
              produce JSON yourself via a system or user message. Without this, the model may
              generate an unending stream of whitespace until the generation reaches the token
              limit, resulting in increased latency and appearance of a "stuck" request. Also
              note that the message content may be partially cut off if
              `finish_reason="length"`, which indicates the generation exceeded `max_tokens`
              or the conversation exceeded the max context length.

          seed: This feature is in Beta. If specified, our system will make a best effort to
              sample deterministically, such that repeated requests with the same `seed` and
              parameters should return the same result. Determinism is not guaranteed, and you
              should refer to the `system_fingerprint` response parameter to monitor changes
              in the backend.

          stop: Up to 4 sequences where the API will stop generating further tokens.

          stream: If set, partial message deltas will be sent, like in ChatGPT. Tokens will be
              sent as data-only
              [server-sent events](https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events/Using_server-sent_events#Event_stream_format)
              as they become available, with the stream terminated by a `data: [DONE]`
              message.
              [Example Python code](https://cookbook.openai.com/examples/how_to_stream_completions).

          temperature: What sampling temperature to use, between 0 and 2. Higher values like 0.8 will
              make the output more random, while lower values like 0.2 will make it more
              focused and deterministic.

              We generally recommend altering this or `top_p` but not both.

          tool_choice: Controls which (if any) function is called by the model. `none` means the model
              will not call a function and instead generates a message. `auto` means the model
              can pick between generating a message or calling a function. Specifying a
              particular function via
              `{"type: "function", "function": {"name": "my_function"}}` forces the model to
              call that function.

              `none` is the default when no functions are present. `auto` is the default if
              functions are present.

          tools: A list of tools the model may call. Currently, only functions are supported as a
              tool. Use this to provide a list of functions the model may generate JSON inputs
              for.

          top_p: An alternative to sampling with temperature, called nucleus sampling, where the
              model considers the results of the tokens with top_p probability mass. So 0.1
              means only the tokens comprising the top 10% probability mass are considered.

              We generally recommend altering this or `temperature` but not both.

          user: A unique identifier representing your end-user, which can help OpenAI to monitor
              and detect abuse.
              [Learn more](https://platform.openai.com/docs/guides/safety-best-practices/end-user-ids).

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        request_body = maybe_transform(
            {
                "messages": messages,
                "model": model,
                "frequency_penalty": frequency_penalty,
                "function_call": function_call,
                "functions": functions,
                "logit_bias": logit_bias,
                "max_tokens": max_tokens,
                "n": n,
                "parallel_tool_calls": parallel_tool_calls,
                "presence_penalty": presence_penalty,
                "response_format": response_format,
                "seed": seed,
                "stop": stop,
                "stream": stream,
                "temperature": temperature,
                "tool_choice": tool_choice,
                "tools": tools,
                "top_p": top_p,
                "user": user,
            },
            completion_create_params.CompletionCreateParams,
        )

        request_body = ChatCompletionsInputValidator(
            request_body=request_body
        ).validate()

        extra_headers = transform_client_header(extra_headers=extra_headers)

        res = self._post(
            "/chat/completions",
            body=request_body,
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
            ),
            cast_to=NexusflowAIChatCompletion,
            stream=stream or False,
            stream_cls=Stream[NexusflowAIChatCompletionChunk],
        )

        if not isinstance(res, Stream):
            maybe_print_hints(res)

            for choice in res.choices:
                parsed = choice.message.parsed
                choice.message.parsed = maybe_parse_into_pydantic(
                    response_format, parsed
                )

        return res

    @required_args(["messages", "model"], ["messages", "model", "stream"])
    def create_with_qwen_agent(
        self,
        *,
        messages: List[ChatCompletionMessageParam],
        model: Union[
            str,
            ALLOWED_MODELS,
        ],
        frequency_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        function_call: completion_create_params.FunctionCall | NotGiven = NOT_GIVEN,
        functions: List[completion_create_params.Function] | NotGiven = NOT_GIVEN,
        logit_bias: Optional[Dict[str, int]] | NotGiven = NOT_GIVEN,
        max_tokens: Optional[int] | NotGiven = NOT_GIVEN,
        n: Optional[int] | NotGiven = NOT_GIVEN,
        parallel_tool_calls: bool | NotGiven = NOT_GIVEN,
        presence_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        response_format: completion_create_params.ResponseFormat | NotGiven = NOT_GIVEN,
        seed: Optional[int] | NotGiven = NOT_GIVEN,
        stop: Union[Optional[str], List[str]] | NotGiven = NOT_GIVEN,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        temperature: Optional[float] | NotGiven = NOT_GIVEN,
        tool_choice: ChatCompletionToolChoiceOptionParam | NotGiven = NOT_GIVEN,
        tools: List[ChatCompletionToolParam] | NotGiven = NOT_GIVEN,
        top_p: Optional[float] | NotGiven = NOT_GIVEN,
        user: str | NotGiven = NOT_GIVEN,
        echo_prompt: str = False,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> NexusflowAIChatCompletion:
        """
        Convert a ChatCompletion with tools or response_format into a Completion with a raw prompt.
        """
        request_body = maybe_transform(
            {
                "messages": messages,
                "model": model,
                "frequency_penalty": frequency_penalty,
                "function_call": function_call,
                "functions": functions,
                "logit_bias": logit_bias,
                "max_tokens": max_tokens,
                "n": n,
                "parallel_tool_calls": parallel_tool_calls,
                "presence_penalty": presence_penalty,
                "response_format": response_format,
                "seed": seed,
                "stop": stop,
                "stream": stream,
                "temperature": temperature,
                "tool_choice": tool_choice,
                "tools": tools,
                "top_p": top_p,
                "user": user,
            },
            completion_create_params.CompletionCreateParams,
        )

        request_body = ChatCompletionsInputValidator(
            request_body=request_body
        ).validate()

        pythonic_request_body = {
            k: v for k, v in request_body.items() if not isinstance(v, NotGiven)
        }

        flow = QwenAgentFlow()
        request_body_to_send = flow.preprocess(pythonic_request_body)
        request_body_to_send = {
            k: NOT_GIVEN if v is None else v for k, v in request_body_to_send.items()
        }

        extra_headers = transform_client_header(extra_headers=extra_headers)

        res = self._post(
            "/chat/completions",
            body=request_body_to_send,
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
            ),
            cast_to=NexusflowAIChatCompletion,
            stream=stream or False,
            stream_cls=Stream[NexusflowAIChatCompletionChunk],
        )

        if not isinstance(res, Stream):
            res = flow.postprocess(pythonic_request_body, res)

            maybe_print_hints(res)

            for choice in res.choices:
                parsed = choice.message.parsed
                choice.message.parsed = maybe_parse_into_pydantic(
                    response_format, parsed
                )

            if echo_prompt:
                res.raw_prompt = request_body_to_send["prompt"]

        return res


class AsyncCompletions(AsyncAPIResource):
    with_raw_response: AsyncCompletionsWithRawResponse

    def __init__(self, client: AsyncNexusflowAI) -> None:
        super().__init__(client)
        self.with_raw_response = AsyncCompletionsWithRawResponse(self)

    @required_args(["messages", "model"], ["messages", "model", "stream"])
    async def create(
        self,
        *,
        messages: List[ChatCompletionMessageParam],
        model: Union[str, ALLOWED_MODELS],
        frequency_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        function_call: completion_create_params.FunctionCall | NotGiven = NOT_GIVEN,
        functions: List[completion_create_params.Function] | NotGiven = NOT_GIVEN,
        logit_bias: Optional[Dict[str, int]] | NotGiven = NOT_GIVEN,
        max_tokens: Optional[int] | NotGiven = NOT_GIVEN,
        n: Optional[int] | NotGiven = NOT_GIVEN,
        parallel_tool_calls: bool | NotGiven = NOT_GIVEN,
        presence_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        response_format: completion_create_params.ResponseFormat | NotGiven = NOT_GIVEN,
        seed: Optional[int] | NotGiven = NOT_GIVEN,
        stop: Union[Optional[str], List[str]] | NotGiven = NOT_GIVEN,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        temperature: Optional[float] | NotGiven = NOT_GIVEN,
        tool_choice: ChatCompletionToolChoiceOptionParam | NotGiven = NOT_GIVEN,
        tools: List[ChatCompletionToolParam] | NotGiven = NOT_GIVEN,
        top_p: Optional[float] | NotGiven = NOT_GIVEN,
        user: str | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> NexusflowAIChatCompletion | AsyncStream[NexusflowAIChatCompletionChunk]:
        """
        Creates a model response for the given chat conversation.

        Args:
          messages: A list of messages comprising the conversation so far.
              [Example Python code](https://cookbook.openai.com/examples/how_to_format_inputs_to_chatgpt_models).

          model: ID of the model to use. See the
              [model endpoint compatibility](https://platform.openai.com/docs/models/model-endpoint-compatibility)
              table for details on which models work with the Chat API.

          frequency_penalty: Number between -2.0 and 2.0. Positive values penalize new tokens based on their
              existing frequency in the text so far, decreasing the model's likelihood to
              repeat the same line verbatim.

              [See more information about frequency and presence penalties.](https://platform.openai.com/docs/guides/gpt/parameter-details)

          function_call: Deprecated in favor of `tool_choice`.

              Controls which (if any) function is called by the model. `none` means the model
              will not call a function and instead generates a message. `auto` means the model
              can pick between generating a message or calling a function. Specifying a
              particular function via `{"name": "my_function"}` forces the model to call that
              function.

              `none` is the default when no functions are present. `auto`` is the default if
              functions are present.

          functions: Deprecated in favor of `tools`.

              A list of functions the model may generate JSON inputs for.

          logit_bias: Modify the likelihood of specified tokens appearing in the completion.

              Accepts a JSON object that maps tokens (specified by their token ID in the
              tokenizer) to an associated bias value from -100 to 100. Mathematically, the
              bias is added to the logits generated by the model prior to sampling. The exact
              effect will vary per model, but values between -1 and 1 should decrease or
              increase likelihood of selection; values like -100 or 100 should result in a ban
              or exclusive selection of the relevant token.

          max_tokens: The maximum number of [tokens](/tokenizer) to generate in the chat completion.

              The total length of input tokens and generated tokens is limited by the model's
              context length.
              [Example Python code](https://cookbook.openai.com/examples/how_to_count_tokens_with_tiktoken)
              for counting tokens.

          n: How many chat completion choices to generate for each input message.

          presence_penalty: Number between -2.0 and 2.0. Positive values penalize new tokens based on
              whether they appear in the text so far, increasing the model's likelihood to
              talk about new topics.

              [See more information about frequency and presence penalties.](https://platform.openai.com/docs/guides/gpt/parameter-details)

          response_format: An object specifying the format that the model must output.

              Setting to `{ "type": "json_object" }` enables JSON mode, which guarantees the
              message the model generates is valid JSON.

              **Important:** when using JSON mode, you **must** also instruct the model to
              produce JSON yourself via a system or user message. Without this, the model may
              generate an unending stream of whitespace until the generation reaches the token
              limit, resulting in increased latency and appearance of a "stuck" request. Also
              note that the message content may be partially cut off if
              `finish_reason="length"`, which indicates the generation exceeded `max_tokens`
              or the conversation exceeded the max context length.

          seed: This feature is in Beta. If specified, our system will make a best effort to
              sample deterministically, such that repeated requests with the same `seed` and
              parameters should return the same result. Determinism is not guaranteed, and you
              should refer to the `system_fingerprint` response parameter to monitor changes
              in the backend.

          stop: Up to 4 sequences where the API will stop generating further tokens.

          stream: If set, partial message deltas will be sent, like in ChatGPT. Tokens will be
              sent as data-only
              [server-sent events](https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events/Using_server-sent_events#Event_stream_format)
              as they become available, with the stream terminated by a `data: [DONE]`
              message.
              [Example Python code](https://cookbook.openai.com/examples/how_to_stream_completions).

          temperature: What sampling temperature to use, between 0 and 2. Higher values like 0.8 will
              make the output more random, while lower values like 0.2 will make it more
              focused and deterministic.

              We generally recommend altering this or `top_p` but not both.

          tool_choice: Controls which (if any) function is called by the model. `none` means the model
              will not call a function and instead generates a message. `auto` means the model
              can pick between generating a message or calling a function. Specifying a
              particular function via
              `{"type: "function", "function": {"name": "my_function"}}` forces the model to
              call that function.

              `none` is the default when no functions are present. `auto` is the default if
              functions are present.

          tools: A list of tools the model may call. Currently, only functions are supported as a
              tool. Use this to provide a list of functions the model may generate JSON inputs
              for.

          top_p: An alternative to sampling with temperature, called nucleus sampling, where the
              model considers the results of the tokens with top_p probability mass. So 0.1
              means only the tokens comprising the top 10% probability mass are considered.

              We generally recommend altering this or `temperature` but not both.

          user: A unique identifier representing your end-user, which can help OpenAI to monitor
              and detect abuse.
              [Learn more](https://platform.openai.com/docs/guides/safety-best-practices/end-user-ids).

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """

        request_body = maybe_transform(
            {
                "messages": messages,
                "model": model,
                "frequency_penalty": frequency_penalty,
                "function_call": function_call,
                "functions": functions,
                "logit_bias": logit_bias,
                "max_tokens": max_tokens,
                "n": n,
                "parallel_tool_calls": parallel_tool_calls,
                "presence_penalty": presence_penalty,
                "response_format": response_format,
                "seed": seed,
                "stop": stop,
                "stream": stream,
                "temperature": temperature,
                "tool_choice": tool_choice,
                "tools": tools,
                "top_p": top_p,
                "user": user,
            },
            completion_create_params.CompletionCreateParams,
        )

        request_body = ChatCompletionsInputValidator(
            request_body=request_body
        ).validate()

        extra_headers = transform_client_header(extra_headers=extra_headers)

        res = await self._post(
            "/chat/completions",
            body=request_body,
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
            ),
            cast_to=NexusflowAIChatCompletion,
            stream=stream or False,
            stream_cls=AsyncStream[NexusflowAIChatCompletionChunk],
        )

        maybe_print_hints(res)

        for choice in res.choices:
            parsed = choice.message.parsed
            choice.message.parsed = maybe_parse_into_pydantic(response_format, parsed)

        return res

    def batch_create(
        self,
        batch: List[ChatCompletionCreateParams],
        num_calls_in_parallel: Optional[int] = None,
    ) -> List[Union[NexusflowAIChatCompletion | Exception]]:
        """
        Will run the batched messages in an async manner
        """
        return batch_call(self.create, batch, num_calls_in_parallel)

    def batch_run_messages(
        self,
        batch_messages: List[List[ChatCompletionMessage]],
        params: ChatCompletionCreateParams,
        num_calls_in_parallel: Optional[int] = None,
    ) -> List[Union[NexusflowAIChatCompletion | Exception]]:
        """
        This function takes in a list of messgaes of type ChatCompletionMessage, appends other parameters (like modelname, temperature, etc) to it and creates a list of ChatCompletionCreateParams objects. These objects are then passed to the batch_create function to be run asynchronously.

        Args:
        batch_messages: A list of messages of type ChatCompletionMessage
        params: A ChatCompletionCreateParams object which is essentially a TypedDict. The key - value pairs should correspond to the input parameters - parameter types of the AsynCompletions.create function.

        Returns:
        A list of NexusflowAIChatCompletion objects.
        """

        if params.get("stream", False):
            raise ValueError("Batch create does not support streaming requests!")

        batched_messages: List[ChatCompletionCreateParams] = [
            {**params, "messages": chat_completion_message}
            for chat_completion_message in batch_messages
        ]

        return self.batch_create(batched_messages, num_calls_in_parallel)

    @required_args(["messages", "model"], ["messages", "model", "stream"])
    def create_with_qwen_agent(
        self,
        *,
        messages: List[ChatCompletionMessageParam],
        model: Union[
            str,
            ALLOWED_MODELS,
        ],
        frequency_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        function_call: completion_create_params.FunctionCall | NotGiven = NOT_GIVEN,
        functions: List[completion_create_params.Function] | NotGiven = NOT_GIVEN,
        logit_bias: Optional[Dict[str, int]] | NotGiven = NOT_GIVEN,
        max_tokens: Optional[int] | NotGiven = NOT_GIVEN,
        n: Optional[int] | NotGiven = NOT_GIVEN,
        parallel_tool_calls: bool | NotGiven = NOT_GIVEN,
        presence_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        response_format: completion_create_params.ResponseFormat | NotGiven = NOT_GIVEN,
        seed: Optional[int] | NotGiven = NOT_GIVEN,
        stop: Union[Optional[str], List[str]] | NotGiven = NOT_GIVEN,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        temperature: Optional[float] | NotGiven = NOT_GIVEN,
        tool_choice: ChatCompletionToolChoiceOptionParam | NotGiven = NOT_GIVEN,
        tools: List[ChatCompletionToolParam] | NotGiven = NOT_GIVEN,
        top_p: Optional[float] | NotGiven = NOT_GIVEN,
        user: str | NotGiven = NOT_GIVEN,
        echo_prompt: str = False,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> NexusflowAIChatCompletion:
        """
        Convert a ChatCompletion with tools or response_format into a Completion with a raw prompt.
        """
        request_body = maybe_transform(
            {
                "messages": messages,
                "model": model,
                "frequency_penalty": frequency_penalty,
                "function_call": function_call,
                "functions": functions,
                "logit_bias": logit_bias,
                "max_tokens": max_tokens,
                "n": n,
                "parallel_tool_calls": parallel_tool_calls,
                "presence_penalty": presence_penalty,
                "response_format": response_format,
                "seed": seed,
                "stop": stop,
                "stream": stream,
                "temperature": temperature,
                "tool_choice": tool_choice,
                "tools": tools,
                "top_p": top_p,
                "user": user,
            },
            completion_create_params.CompletionCreateParams,
        )

        request_body = ChatCompletionsInputValidator(
            request_body=request_body
        ).validate()

        pythonic_request_body = {
            k: v for k, v in request_body.items() if not isinstance(v, NotGiven)
        }

        flow = QwenAgentFlow()
        request_body_to_send = flow.preprocess(pythonic_request_body)
        request_body_to_send = {
            k: NOT_GIVEN if v is None else v for k, v in request_body_to_send.items()
        }

        extra_headers = transform_client_header(extra_headers=extra_headers)

        res = self._post(
            "/chat/completions",
            body=request_body_to_send,
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
            ),
            cast_to=NexusflowAIChatCompletion,
            stream=stream or False,
            stream_cls=Stream[NexusflowAIChatCompletionChunk],
        )

        if not isinstance(res, Stream):
            res = flow.postprocess(pythonic_request_body, res)

            maybe_print_hints(res)

            for choice in res.choices:
                parsed = choice.message.parsed
                choice.message.parsed = maybe_parse_into_pydantic(
                    response_format, parsed
                )

            if echo_prompt:
                res.raw_prompt = request_body_to_send["prompt"]

        return res


class CompletionsWithRawResponse:
    def __init__(self, completions: Completions) -> None:
        self.create = to_raw_response_wrapper(
            completions.create,
        )


class AsyncCompletionsWithRawResponse:
    def __init__(self, completions: AsyncCompletions) -> None:
        self.create = async_to_raw_response_wrapper(
            completions.create,
        )



class DeprecationError(Exception):
    """An exception that indicates an error when using deprecated params"""


class NexusflowAIError(Exception):
    """An exception that indicates an error which creating the NexusflowAI client"""



from typing import List, Any, Dict, Union, Optional

from dataclasses import dataclass

import json

from openai.types.chat import ChatCompletionToolParam
from openai.types.chat.chat_completion_tool_choice_option_param import (
    ChatCompletionNamedToolChoiceParam,
)

from qwen_agent.llm.schema import Message, FunctionCall, ContentItem
from qwen_agent.llm.fncall_prompts.qwen_fncall_prompt import FN_STOP_WORDS

from nexusflowai.types import (
    NexusflowAIChatCompletion,
    NexusflowAIChatCompletionMessage,
)
from nexusflowai.validators.base_input_validator import BOT_END

from nexusflowai.validators.create_with_tools.postprocess import (
    FunctionCallResponseTranslator,
    RavenFunctionCall,
)


@dataclass
class QwenMessageTranslator:
    def convert_request_to_qwen_params(
        self,
        request_dict: NexusflowAIChatCompletion,
    ) -> Dict[str, Any]:
        qwen_messages = []
        # pylint: disable=too-many-nested-blocks
        for turn, m in enumerate(request_dict["messages"]):
            if m["role"] == "tool":
                content = self._format_content_parts(m["content"])
                msg = {
                    "role": "function",
                    "content": content,
                }
                assert "tool_call_id" in m
                f = False
                for prev_turn in reversed(range(turn)):
                    if f:
                        break
                    if (
                        request_dict["messages"][prev_turn]["role"] == "assistant"
                        and "tool_calls" in request_dict["messages"][prev_turn]
                        and request_dict["messages"][prev_turn]["tool_calls"]
                        is not None
                    ):
                        for t in request_dict["messages"][prev_turn]["tool_calls"]:
                            if f:
                                break
                            if m["tool_call_id"] == t["id"]:
                                msg["name"] = t["function"]["name"]
                                f = True
                                break
                qwen_messages.append(Message(**msg))
            elif (
                m["role"] == "assistant"
                and "tool_calls" in m
                and m["tool_calls"] is not None
            ):
                for t in m["tool_calls"]:
                    content = self._format_content_parts(m["content"])
                    qwen_messages.append(
                        Message(
                            role=m["role"],
                            content=content,
                            function_call=FunctionCall(
                                name=t["function"]["name"],
                                arguments=t["function"]["arguments"],
                            ),
                        )
                    )
            else:
                content = self._format_content_parts(m["content"])
                qwen_messages.append(Message(role=m["role"], content=content))
        qwen_functions = (
            [self._format_function(t["function"]) for t in request_dict["tools"]]
            if request_dict.get("tools") is not None
            else None
        )

        request_dict["messages"] = qwen_messages
        request_dict["tools"] = qwen_functions

        return request_dict

    def _format_content_parts(
        self, content: Optional[Union[str, List[Dict[str, Any]]]]
    ) -> List[ContentItem]:
        """Format content str or ContentPartParams"""
        if not content:
            return []
        elif isinstance(content, str):
            return [ContentItem(text=content)]
        elif isinstance(content, list):
            contents = []
            for content_part in content:
                if "text" not in content_part:
                    raise NotImplementedError
                contents.append(ContentItem(text=content_part["text"]))
            return contents
        else:
            raise NotImplementedError

    def _format_function(self, function_dict: Dict[str, Any]) -> Dict[str, Any]:
        """Format functions to be presentable to Qwen-Agent."""

        if "description" not in function_dict:
            # https://github.com/QwenLM/Qwen-Agent/blob/8fe8f09c2f43f768f334e7a9244b64d89fcb088b/qwen_agent/llm/fncall_prompts/qwen_fncall_prompt.py#L348
            function_dict["description"] = ""

        return function_dict

    def convert_qwen_messages_to_openai_message(
        self,
        messages: List[Message],
        tools: List[ChatCompletionToolParam],
        tool_choice: ChatCompletionNamedToolChoiceParam | None = None,
    ) -> NexusflowAIChatCompletionMessage:
        fcrt = FunctionCallResponseTranslator()

        contents = []

        tool_calls = []
        for message in messages:
            message_content = message.content
            if isinstance(message_content, list):
                message_content = "\n".join([m.text for m in message_content])

            function_call = message.function_call
            if function_call:
                message_content = ""

                fc_dict = function_call.model_dump(mode="json")
                try:
                    raven_fc = RavenFunctionCall.from_tool_call({"function": fc_dict})
                    raven_tool_calls, _, _ = fcrt.raw_response_to_tool_calls(
                        tools=tools,
                        raw_response=repr(raven_fc),
                        tool_choice=tool_choice,
                    )
                except:  # pylint: disable=bare-except
                    raven_fc = None
                    raven_tool_calls = None

                # We only take the first item since we know it's only one tool call.
                if raven_tool_calls:
                    tool_calls.append(raven_tool_calls[0])
                elif raven_fc:
                    fc_dict_to_output = {
                        "name": raven_fc.name,
                        "arguments": json.dumps(raven_fc.kwargs),
                    }
                    tool_calls.append(fcrt.create_tool_call(fc_dict_to_output))
                else:
                    fc_dict_to_output = function_call.model_dump(mode="json")
                    tool_calls.append(fcrt.create_tool_call(fc_dict_to_output))

            contents.append(message_content)

        tool_calls = tool_calls or None
        content = "".join(contents) or None

        return NexusflowAIChatCompletionMessage(
            content=content, role="assistant", tool_calls=tool_calls
        )

    def adjust_stop_words_for_qwen_agent(self, stop: List[str]) -> List[str]:
        if BOT_END in stop:
            stop.remove(BOT_END)
        stop = stop + [x for x in FN_STOP_WORDS if x not in stop]
        return stop



from typing import List, Optional

from openai.types.chat.chat_completion_chunk import (
    ChatCompletionChunk as BaseChatCompletionChunk,
)

__all__ = ["NexusflowAIChatCompletionChunk"]


class NexusflowAIChatCompletionChunk(BaseChatCompletionChunk):
    hints: Optional[List[str]] = None



from __future__ import annotations

from typing import Any, Callable, Dict, List, Optional, Union

from logging import ERROR, getLogger

from asyncio import Semaphore, run
from tqdm.asyncio import tqdm

from pydantic import BaseModel

from openai._types import NOT_GIVEN, NotGiven
from openai.types import CompletionCreateParams
from openai.types.chat.completion_create_params import (
    CompletionCreateParams as ChatCompletionCreateParams,
)

from nexusflowai import logger
from nexusflowai.types import (
    NexusflowAICompletion,
    NexusflowAIChatCompletion,
    NexusflowAIChatCompletionChunk,
)


def maybe_print_hints(
    res: NexusflowAIChatCompletion | NexusflowAIChatCompletionChunk,
) -> None:
    hints = res.hints or []
    if hints:
        logger.warning("\n".join(hints))


def maybe_parse_into_pydantic(
    response_format: Union[BaseModel, Dict[str, Any], NotGiven],
    parsed: Dict[str, Any] | None,
) -> None:
    if response_format is NOT_GIVEN or parsed is None:
        return parsed

    is_pydantic = isinstance(response_format, type) and issubclass(
        response_format, BaseModel
    )
    if not is_pydantic:
        return parsed

    return response_format.model_validate(parsed)


def batch_call(
    caller: Callable,
    batch: List[CompletionCreateParams, ChatCompletionCreateParams],
    num_calls_in_parallel: Optional[int],
) -> List[Union[NexusflowAICompletion, NexusflowAIChatCompletion | Exception]]:
    assert not any(
        params.get("stream", False) for params in batch
    ), "Batch create does not support streaming requests!"

    semaphore = None
    if num_calls_in_parallel is not None:
        semaphore = Semaphore(num_calls_in_parallel)

    async def create_with_except(**params):
        try:
            if semaphore is not None:
                async with semaphore:
                    return await caller(**params)
            else:
                return await caller(**params)
        except Exception as e:  # pylint: disable=broad-exception-caught
            # We can handle basic exceptions
            if len(e.args) == 1:
                print(f"{type(e).__name__}: {e.args[0]}")
            else:
                raise e
            return e

    async def gather_responses():
        return await tqdm.gather(*(create_with_except(**params) for params in batch))

    # Set the log level higher to avoid so many prints
    logger_names = ["openai._base_client", "httpx"]
    loggers = list(map(getLogger, logger_names))
    original_log_levels = [l.level for l in loggers]
    for l in loggers:
        l.setLevel(ERROR)

    res = run(gather_responses())

    for l, original_log_level in zip(loggers, original_log_levels):
        l.setLevel(original_log_level)

    return res



from typing import Any, Dict, List, Optional, Type, Union

from dataclasses import dataclass

from collections import Counter

from copy import deepcopy

import json
from json import JSONDecodeError

import json_repair

from openai.types.chat import ChatCompletionMessageToolCallParam
from openai.types.chat.completion_create_params import (
    CompletionCreateParams as ChatCompletionCreateParams,
    ChatCompletionMessageParam,
)
from openai.types.chat import ChatCompletionToolParam

from nexusflowai.validators.json_schema_to_dataclasses import (
    try_convert_to_dataclasses_str,
)
from nexusflowai.validators.create_with_tools.utils import (
    ALLOWED_ROLES,
    CleanedFunction,
    get_tool_choice_prefix,
    resolve_tools_and_tool_choice,
    _parse_content_parts,
)


@dataclass
class RavenType:
    TYPE_KEY = "type"

    schema: Any

    @classmethod
    def format_type_for_prompt(
        cls,
        parameter_type: Union[str, List[Any]],
        parameter_schema: Optional[Dict[str, Any]] = None,
    ) -> Optional[str]:
        """Format the specified parameter (or return) type.

        Return a string representation of the specified parameter type that is to
        be used in a prompt.
        """

        if isinstance(parameter_type, list):
            alternative_types = Counter()
            for type_name in parameter_type:
                if not isinstance(type_name, str):
                    continue

                alternative_type = cls.format_type_for_prompt(type_name)
                if alternative_type is not None:
                    alternative_types[alternative_type] += 1

            num_alternative_types = len(alternative_types)
            if num_alternative_types > 1:
                types_string = ", ".join(alternative_types)
                return f"Union[{types_string}]"
            elif num_alternative_types == 1:
                return next(iter(alternative_types))
            else:
                return None

        if parameter_type == "string":
            return "str"

        if parameter_type == "number":
            return "float"

        if parameter_type == "integer":
            return "int"

        # pylint: disable=too-many-nested-blocks
        if parameter_type == "object":
            # In principle, a type could be assigned to the values in a
            # dictionary based on the types of the properties of the object.
            # However, this could be complicated by the possibility that there
            # could be additional properties in an object that are not
            # specified in the properties.  For now, no types are provided for
            # the keys and values in the dictionary for simplicity.

            # "Dict" for (nested) dict.
            if parameter_schema is not None:
                properties = parameter_schema.get("properties")
                if isinstance(properties, dict):
                    for _prop_key, prop_value in properties.items():
                        if not isinstance(prop_value, dict):
                            continue
                        element_type = prop_value.get(cls.TYPE_KEY)
                        if isinstance(element_type, (str, list)):
                            prompt_element_type = cls.format_type_for_prompt(
                                element_type, parameter_schema=prop_value
                            )
                            if prompt_element_type is not None:
                                return f"Dict[str, {prompt_element_type}]"
                    return "Dict"

                extra_properties = parameter_schema.get("additionalProperties")
                if isinstance(extra_properties, dict):
                    element_type = extra_properties.get(cls.TYPE_KEY)
                    if isinstance(element_type, (str, list)):
                        prompt_element_type = cls.format_type_for_prompt(
                            element_type, parameter_schema=extra_properties
                        )
                        if prompt_element_type is not None:
                            return f"Dict[str, {prompt_element_type}]"
                    return "Dict"

            # Otherwise, prefer "object" instead of "dict".
            return "object"

        if parameter_type == "array":
            if parameter_schema is not None:
                array_items = parameter_schema.get("items")
                if isinstance(array_items, dict):
                    element_type = array_items.get(cls.TYPE_KEY)
                    if isinstance(element_type, (str, list)):
                        prompt_element_type = cls.format_type_for_prompt(
                            element_type, parameter_schema=array_items
                        )
                        if prompt_element_type is not None:
                            return f"List[{prompt_element_type}]"

            return "list"

        if parameter_type == "boolean":
            return "bool"

        if parameter_type == "null":
            return "None"

        return None


@dataclass
class RavenFunctionReturn(RavenType):
    """This class represents a function return type."""

    def format_for_prompt_definition(self) -> Union[str, None]:
        """Format this function return type.

        Return a string representation of this function return type that is to be
        part of the function definition of a prompt.
        """

        if not isinstance(self.schema, dict):
            return None

        parameter_type = self.schema.get(self.TYPE_KEY)
        if not isinstance(parameter_type, (str, list)):
            return None

        prompt_type = self.format_type_for_prompt(
            parameter_type, parameter_schema=self.schema
        )
        if not prompt_type:
            return None
        else:
            return f"{prompt_type}"


@dataclass
class RavenFunctionParameter(RavenType):
    """This class represents a parameter in a function definition."""

    DESC_KEY = "description"

    name: str
    required: bool = True

    def format_for_prompt_docstring(self) -> Union[str, None]:
        """Format this function parameter.

        Return a string representation of this function parameter that is to be
        described in the function docstring of a prompt.
        """

        if not isinstance(self.schema, dict):
            return None

        parameter_desc = self.schema.get(self.DESC_KEY)
        if isinstance(parameter_desc, str):
            parameter_desc = parameter_desc.strip()
        else:
            if "items" in self.schema and "description" in self.schema["items"]:
                parameter_desc = self.schema["items"]["description"]
                if isinstance(parameter_desc, str):
                    parameter_desc = parameter_desc.strip()
                else:
                    parameter_desc = ""
            else:
                parameter_desc = ""
        if parameter_desc:
            # Parameter description begins with upper case.
            parameter_desc = f"{parameter_desc[0].upper()}{parameter_desc[1:]}"

        parameter_type = self.schema.get(self.TYPE_KEY)
        if not isinstance(parameter_type, (str, list)):
            return None

        prompt_type = self.format_type_for_prompt(
            parameter_type, parameter_schema=self.schema
        )

        desc_parts = [parameter_desc.strip()]

        # Parameter descriptions end with a dot.
        if desc_parts[0] and not desc_parts[0].endswith("."):
            desc_parts.append(".")

        # "robustly" handling default values.
        if "default" in self.schema:
            default = self.schema.get("default")
            if default is not None and parameter_type is not None:
                parameter_cls_name = self.format_type_for_prompt(parameter_type, None)
                parameter_cls = eval(parameter_cls_name)  # pylint: disable=eval-used
                if not (parameter_cls is object or issubclass(parameter_cls, str)):
                    try:
                        if isinstance(default, str):
                            default = parameter_cls(json.loads(default))
                        else:
                            default = parameter_cls(default)
                    except Exception:  # pylint: disable=broad-exception-caught
                        if default == "None":
                            default = None
            desc_parts.append(f" Default value: {repr(default)}.")
        elif not self.required:
            # Sometimes the description specifies a default value using natural language
            # but that same default value does not occur in the schema.
            default = None
            desc_parts.append(f" Default value: {repr(default)}.")

        # Optional parameters.
        if not self.required:
            desc_parts.append(" (Optional)")

        # Nested dicts.
        if "properties" in self.schema and isinstance(self.schema["properties"], dict):
            desc_parts.append(" Has keys:")
            some_prop_doc = False
            for prop_key, prop_value in self.schema["properties"].items():
                prop_type = prop_value.get(self.TYPE_KEY)
                if not isinstance(prop_type, str):
                    continue
                some_prop_doc = True
                prop_prompt_type = self.format_type_for_prompt(
                    prop_type, parameter_schema=prop_value
                )
                prop_desc = prop_value.get(self.DESC_KEY)
                if isinstance(prop_desc, str):
                    prop_desc = prop_desc.strip()
                if not prop_prompt_type:
                    if not prop_desc:
                        desc_parts.append(f"\n    - {prop_key}")
                    else:
                        desc_parts.append(f"\n    - {prop_key}: {prop_desc}")
                else:
                    if not prop_desc:
                        desc_parts.append(f"\n    - {prop_key} ({prop_prompt_type})")
                    else:
                        desc_parts.append(
                            f"\n    - {prop_key} ({prop_prompt_type}): {prop_desc}"
                        )
            if not some_prop_doc:
                desc_parts.append("\n")

        # Not duplicating enum lists in parameter description.
        min_enum_occur_ct = 3
        if "enum" in self.schema and isinstance(self.schema["enum"], list):
            occur_ct = 0
            for value in self.schema["enum"]:
                if parameter_desc.find(value) >= 0:
                    occur_ct += 1
                    if occur_ct >= min(min_enum_occur_ct, len(self.schema["enum"])):
                        break
            if not occur_ct >= min(min_enum_occur_ct, len(self.schema["enum"])):
                desc_parts.append(" Available values:")
                for value in self.schema["enum"]:
                    desc_parts.append(f"\n    - {repr(value)}")
        elif (
            isinstance(prompt_type, str)
            and prompt_type.startswith("List[")
            and "enum" in self.schema["items"]
            and isinstance(self.schema["items"]["enum"], list)
        ):
            occur_ct = 0
            for value in self.schema["items"]["enum"]:
                if parameter_desc.find(value) >= 0:
                    occur_ct += 1
                    if occur_ct >= min(
                        min_enum_occur_ct, len(self.schema["items"]["enum"])
                    ):
                        break
            if not (
                occur_ct >= min(min_enum_occur_ct, len(self.schema["items"]["enum"]))
            ):
                desc_parts.append(" Available entries:")
                for value in self.schema["items"]["enum"]:
                    desc_parts.append(f"\n    - {repr(value)}")

        parameter_desc = "".join(desc_parts)

        if not prompt_type:
            if not parameter_desc:
                return f"- {self.name}"
            else:
                return f"- {self.name}: {parameter_desc}"
        else:
            if not parameter_desc:
                return f"- {self.name} ({prompt_type})"
            else:
                return f"- {self.name} ({prompt_type}): {parameter_desc}"

    def format_for_prompt_definition(self) -> str:
        """Format this function parameter.

        Return a string representation of this function parameter that is to be
        part of the function definition of a prompt.
        """

        if not isinstance(self.schema, dict):
            if self.required:
                return self.name
            else:
                default = None
                return f"{self.name}={repr(default)}"

        parameter_type = self.schema.get(self.TYPE_KEY)
        default = self.schema.get("default")

        # "robustly" handling default values.
        if "default" in self.schema:
            if default is not None and parameter_type is not None:
                parameter_cls_name = self.format_type_for_prompt(parameter_type, None)
                parameter_cls = eval(parameter_cls_name)  # pylint: disable=eval-used
                if not (parameter_cls is object or issubclass(parameter_cls, str)):
                    try:
                        if isinstance(default, str):
                            default = parameter_cls(json.loads(default))
                        else:
                            default = parameter_cls(default)
                    except Exception:  # pylint: disable=broad-exception-caught
                        if default == "None":
                            default = None

        if not isinstance(parameter_type, (str, list)):
            if self.required:
                return self.name
            else:
                return f"{self.name}={repr(default)}"

        prompt_type = self.format_type_for_prompt(
            parameter_type, parameter_schema=self.schema
        )
        # Required param with default value.
        if not prompt_type:
            if not self.required or "default" in self.schema:
                return f"{self.name}={repr(default)}"
            else:
                return self.name
        else:
            if not self.required or "default" in self.schema:
                return f"{self.name}: {prompt_type} = {repr(default)}"
            else:
                return f"{self.name}: {prompt_type}"


@dataclass
class RavenFunctionDefinition:
    """This class represents a function definition."""

    NEWLINE = "\n"
    DOCSTRING_INDENT = "    "
    DOCSTRING_TEMPLATE = '''{indent}"""
{indent}{description}"""
'''
    FUNCTION_TEMPLATE = """Function:
def {name}({parameters}){returns}:
{docstring}
"""

    name: str
    parameters: List[RavenFunctionParameter]  # non-standard "returns"
    returns: List[RavenFunctionReturn]
    description: Optional[str]  # non-standard "extra_description"
    extra_description: Optional[str] = None

    def format_for_prompt(self) -> str:
        """Format this function definition.

        Return a string representation of this function definition that is to be
        used in a prompt.
        """

        parameter_strings = [
            parameter.format_for_prompt_definition() for parameter in self.parameters
        ]
        combined_parameters = ", ".join(parameter_strings)

        if not self.returns:
            returns = ""
        else:
            returns = self.returns[0].format_for_prompt_definition()
            if returns is None:
                returns = ""
            else:
                returns = f" -> {returns}"

        if self.description is not None:
            desc_blocks = [self.description.strip()]
            for p in self.parameters:
                d = p.format_for_prompt_docstring()
                if d is not None:
                    if len(desc_blocks) <= 1:
                        desc_blocks.append("")
                        desc_blocks.append("Parameters:")
                    desc_blocks.append(d)
            if self.extra_description is not None:
                desc_blocks.append("")
                desc_blocks.append(self.extra_description.strip())
            formatted_lines = []
            formatted_lines.append(f'{self.DOCSTRING_INDENT}"""')
            for block in desc_blocks:
                if block:
                    for line in block.splitlines():
                        formatted_lines.append(f"{self.DOCSTRING_INDENT}{line}")
                else:
                    formatted_lines.append("")
            formatted_lines.append(f'{self.DOCSTRING_INDENT}"""')
            formatted_lines.append("")
            docstring = self.NEWLINE.join(formatted_lines)
        else:
            docstring = ""

        return self.FUNCTION_TEMPLATE.format(
            name=self.name,
            parameters=combined_parameters,
            returns=returns,
            docstring=docstring,
        )


@dataclass
class ToolsDefs:
    """A tool with $defs"""

    tool: ChatCompletionToolParam

    def format_for_prompt(self, with_import=False) -> str:
        tool_params = self.tool["function"]["parameters"]
        _, dataclasses_str, _ = try_convert_to_dataclasses_str(
            tool_params, with_import=with_import
        )
        dataclasses_str = f"{dataclasses_str}\n\n"
        return dataclasses_str


@dataclass
class RavenFunctionCall:
    name: str
    kwargs: Dict[str, Any]

    def __repr__(self) -> str:
        kwargs_repr = ", ".join(f"{k}={repr(v)}" for k, v in self.kwargs.items())
        return f"{self.name}({kwargs_repr})"

    @classmethod
    def from_tool_call(cls, tool_call: Dict[str, Any]) -> "RavenFunctionCall":
        function_name = tool_call["function"]["name"]
        function_kwargs = json_repair.loads(tool_call["function"]["arguments"])
        return cls(
            name=function_name,
            kwargs=function_kwargs,
        )


@dataclass
class RavenQuery:
    """This class represents a query that is to be sent to the Raven model."""

    PROMPT_TEMPLATE = "{functions}{response_format}{turns}{tool_choice_prefix}"
    HUMAN_END_SUFFIX = " <human_end>\n\nCall:"

    tools_list: List[Union[RavenFunctionDefinition, ToolsDefs]]
    response_format: str
    messages: List[Any]
    tool_choice_prefix: str | None

    def construct_prompt(self) -> str:
        """Return the prompt for the Raven model for this query."""

        function_strings = [tool.format_for_prompt() for tool in self.tools_list]
        combined_functions = "".join(function_strings)

        response_format = self.response_format
        if self.response_format is None:
            response_format = ""
        elif isinstance(response_format, dict):
            _, dataclasses_str, _ = try_convert_to_dataclasses_str(response_format)
            """
            @dataclass
            class Person:
                name: str
                age: int
                email: Optional[str] = None
            """
            # pylint: disable=use-maxsplit-arg
            top_level_obj_cls = dataclasses_str.split("@dataclass\nclass ")[-1]
            top_level_obj_cls = top_level_obj_cls.split(":")[0]

            indent = RavenFunctionDefinition.DOCSTRING_INDENT
            response_format = f"""{dataclasses_str}
Function:
def extract_item(value: {top_level_obj_cls}): 
{indent}\"\"\"
{indent}Please extract the necessary information in the format requested using the data-classes above and pass them in.
{indent}\"\"\"

"""
            combined_functions = ""
        else:
            raise NotImplementedError("Unknown response format schema")

        turns = self.format_turns(self.messages)

        tool_choice_prefix = ""
        if self.tool_choice_prefix is not None:
            tool_choice_prefix = f" {self.tool_choice_prefix}"

        return self.PROMPT_TEMPLATE.format(
            functions=combined_functions,
            response_format=response_format,
            turns=turns,
            tool_choice_prefix=tool_choice_prefix,
        )

    @staticmethod
    def construct_tool_defs_strs(
        tool_defs: List[ChatCompletionToolParam], with_import: bool = False
    ) -> List[str]:
        if not tool_defs:
            return []
        else:
            dataclasses_strs = []
            for tool in tool_defs:
                tool_params = tool["function"]["parameters"]
                _, dataclasses_str, _ = try_convert_to_dataclasses_str(
                    tool_params, with_import=with_import
                )
                dataclasses_strs.append(dataclasses_str)
        return dataclasses_strs

    def create_initial_system_instructions(
        self, messages: List[ChatCompletionMessageParam]
    ) -> str:
        initial_system_instructions = ""

        initial_system_queries = []
        for message in messages:
            if message["role"] != "system":
                break

            system_query = message["content"]
            if isinstance(system_query, List):
                content_strs = _parse_content_parts(system_query)
                system_query = "\n".join(content_strs)
            initial_system_queries.append(system_query)

        system_query = "\n".join(initial_system_queries)
        system_query = system_query or None

        if system_query:
            initial_system_instructions = f"Instruction: {system_query.strip()}\n\n"

        return initial_system_instructions

    def format_turns(self, messages: List[ChatCompletionMessageParam]) -> str:
        user_query_idxs = [i for i, m in enumerate(messages) if m["role"] == "user"]

        shifted_queries = messages.copy()
        # Remove initial system queries
        while len(shifted_queries) and shifted_queries[0]["role"] == "system":
            shifted_queries.pop(0)

        non_user_query_idxs = [
            i for i, m in enumerate(shifted_queries) if m["role"] != "user"
        ]
        # Shift all non-user queries to come before user queries
        for idx in non_user_query_idxs:
            if idx != 0:
                shifted_queries[idx - 1], shifted_queries[idx] = (
                    shifted_queries[idx],
                    shifted_queries[idx - 1],
                )

        tool_call_id_to_result = dict()
        input_tool_param_ids = []
        for message in messages:
            if message["role"] != "tool":
                continue

            try:
                content = json.loads(message["content"])
            except JSONDecodeError:
                content = message["content"]

            tool_call_id_to_result[message["tool_call_id"]] = content

            input_tool_param_ids.append(message["tool_call_id"])

        input_tool_call_param_ids = []
        for message in messages:
            if message["role"] != "assistant":
                continue

            input_tool_call_param_ids.extend(
                [t["id"] for t in message.get("tool_calls") or []]
            )

        # We use lists here rather than sets to keep the order of the returns the same
        # as the order of the inputs. This is more intuitive to the user since there is an ordering.
        extra_tool_results = [
            i for i in input_tool_param_ids if i not in input_tool_call_param_ids
        ]
        assert not extra_tool_results, f"""ToolMessage / ToolCallParam mismatch! (extra ChatCompletionToolMessageParam `tool_call_id`s)
The following ChatCompletionToolMessageParam `tool_call_id`s don't have a corresponding ChatCompletionAssistantMessageParam ChatCompletionMessageToolCallParam `id`.
Extra ChatCompletionToolMessageParam `tool_call_id`s: {extra_tool_results}
Here are the ChatCompletionAssistantMessageParam ChatCompletionMessageToolCallParam `id`s that we found: {input_tool_call_param_ids}"""

        num_turns = len(user_query_idxs)

        res = self.create_initial_system_instructions(messages)
        is_new_turn = True
        is_new_assistant_turn = True
        turn = 1
        for message in shifted_queries:
            is_last_turn = turn >= num_turns

            message_content = message.get("content", [])
            if message_content:
                if isinstance(message_content, List):
                    content_strs = _parse_content_parts(message_content)
                else:
                    content_strs = [message_content]
            else:
                content_strs = []

            if is_new_turn:
                if is_last_turn:
                    res += "Current Turn:\n"
                else:
                    res += f"Turn {turn}:\n"

            if message["role"] == "assistant":
                tool_calls = message.get("tool_calls") or []

                # Chat Response
                for content in content_strs:
                    if is_last_turn:
                        res += f"Original Plan: {content}\n"
                    else:
                        res += f"Previous Original Plan: {content}\n"

                if is_new_assistant_turn:
                    # Planning
                    previous_original_plan = json.loads(
                        message.get("nexusflowai_extras") or "{}"
                    ).get("original_plan")
                    if previous_original_plan is not None:
                        if is_last_turn:
                            res += f"Original Plan: {previous_original_plan}\n"
                        else:
                            res += f"Previous Original Plan: {previous_original_plan}\n"

                for tool_call in tool_calls:
                    tool_call_id = tool_call["id"]
                    if tool_call_id not in tool_call_id_to_result:
                        continue

                    previous_call = RavenFunctionCall.from_tool_call(tool_call)
                    previous_result = tool_call_id_to_result.pop(tool_call_id)

                    res += f"""Previous Call: {previous_call}
Previous Result: {previous_result}
"""
                is_new_turn = False
                is_new_assistant_turn = False

            elif message["role"] == "system":
                for content in content_strs:
                    res += f"Instruction: {content}\n"
                is_new_turn = False

            elif message["role"] == "user":
                if is_last_turn:
                    res += self._format_final_user_query(content_strs)
                else:
                    for content in content_strs[:-1]:
                        res += f"Previous User Query: {content}\n"
                    res += f"Previous User Query: {content_strs[-1]} <turn_end>\n\n"

                turn += 1
                is_new_turn = True
                is_new_assistant_turn = True
            else:
                is_new_turn = False

        # Sometimes only system prompts are input
        if len(user_query_idxs) == 0:
            res += self.HUMAN_END_SUFFIX

        return res

    def _format_final_user_query(self, content_strs: List[str]) -> str:
        final_query_str = "\n".join(
            [f"User Query: {content}\n" for content in content_strs[:-1]]
        )
        final_query_str += f"User Query: {content_strs[-1]}{self.HUMAN_END_SUFFIX}"
        return final_query_str


class FunctionCallTranslator:
    RETURNS_KEY = "returns"  # Non-standard "returns".
    EXTRA_DESCRIPTION_KEY = "extra_description"  # Non-standard "extra_description".

    RAVEN_QUERY_CLS: Type[RavenQuery] = RavenQuery

    def get_raven_func_definitions(
        self, tools: List[Dict[str, Any]]
    ) -> List[RavenFunctionDefinition]:
        functions = []
        for tool in tools:
            function_definition = self.get_raven_fun_definition(tool)
            functions.append(function_definition)
        return functions

    def get_raven_fun_definition(
        self, tool: ChatCompletionToolParam
    ) -> RavenFunctionDefinition:
        tool_function = tool["function"]
        function_name = tool_function["name"]
        function_parameters = tool_function["parameters"]

        defined_parameters = []
        required_parameters = set(function_parameters.get("required", []))
        function_properties = function_parameters.get("properties", {})
        for parameter_name, parameter_schema in function_properties.items():
            defined_parameter = RavenFunctionParameter(
                name=parameter_name,
                schema=parameter_schema,
                required=parameter_name in required_parameters,
            )
            defined_parameters.append(defined_parameter)

        if self.RETURNS_KEY not in tool_function:
            function_returns = []
        else:
            function_returns = [
                RavenFunctionReturn(schema=tool_function[self.RETURNS_KEY])
            ]

        function_description = tool_function.get("description")

        function_extra_description = tool_function.get(self.EXTRA_DESCRIPTION_KEY)

        function_definition = RavenFunctionDefinition(
            name=function_name,
            parameters=defined_parameters,
            returns=function_returns,
            description=function_description,
            extra_description=function_extra_description,
        )
        return function_definition

    def format_tools(
        self, tools: List[ChatCompletionToolParam]
    ) -> List[Union[RavenFunctionDefinition, ToolsDefs]]:
        tools_list = []
        for tool in tools:
            if "$defs" in tool.get("function", {}).get("parameters", {}):
                tools_list.append(ToolsDefs(tool))
            else:
                tools_list.append(self.get_raven_fun_definition(tool))
        return tools_list

    def create_params_to_fc_prompt(
        self, create_params: ChatCompletionCreateParams
    ) -> str:
        messages = create_params["messages"]

        for message in messages:
            role = message["role"]
            assert (
                role in ALLOWED_ROLES
            ), f"The `{role}` role is not supported. Supported roles: {repr(ALLOWED_ROLES)}"

        tools = create_params.get("tools") or []
        tools, tool_choice = resolve_tools_and_tool_choice(
            tools, create_params.get("tool_choice")
        )
        cfs = list(map(CleanedFunction.from_tool_param, tools))
        dirty_function_name_to_cf = {cf.dirty_name: cf for cf in cfs}
        tool_choice_prefix = get_tool_choice_prefix(
            tool_choice, dirty_function_name_to_cf
        )
        tools = [cf.clean_tool_param(tool) for cf, tool in zip(cfs, tools)]

        messages = deepcopy(messages)
        for message in messages:
            if message["role"] != "assistant":
                continue

            tool_calls = message.get("tool_calls")
            if not tool_calls:
                continue

            clean_tool_calls = []
            for dirty_tool_call in tool_calls:
                dirty_tool_call: ChatCompletionMessageToolCallParam

                cf = dirty_function_name_to_cf.get(dirty_tool_call["function"]["name"])
                if cf is None:
                    clean_tool_calls.append(dirty_tool_call)
                else:
                    clean_tool_call = cf.clean_tool_call_param(dirty_tool_call)
                    clean_tool_calls.append(clean_tool_call)

            message["tool_calls"] = clean_tool_calls

        tools_list = self.format_tools(tools)

        response_format = create_params.get("response_format")

        query = self.RAVEN_QUERY_CLS(
            tools_list=tools_list,
            response_format=response_format,
            messages=messages,
            tool_choice_prefix=tool_choice_prefix,
        )
        return query.construct_prompt()



from copy import deepcopy

from dataclasses import dataclass

import json
from json import JSONDecodeError

import json_repair

from qwen_agent.llm.base import _postprocess_stop_words
from qwen_agent.llm.fncall_prompts.qwen_fncall_prompt import QwenFnCallPrompt
from qwen_agent.llm.function_calling import (
    validate_num_fncall_results,
    simulate_response_completion_with_chat,
)
from qwen_agent.llm.schema import Message, ContentItem, DEFAULT_SYSTEM_MESSAGE
from qwen_agent.utils.utils import format_as_text_message

from nexusflowai.types import (
    NexusflowAIChatCompletion,
    NexusflowAIChatCompletionCreateParams,
)
from nexusflowai.types.chat_completion import Choice as NexusflowAIChatCompletionChoice

from nexusflowai.validators.qwen_agent.qwen_message_translator import (
    QwenMessageTranslator,
)


@dataclass
class QwenAgentFlow:
    fncall_prompt = QwenFnCallPrompt()
    qwen_translator = QwenMessageTranslator()

    def preprocess(
        self, create_params: NexusflowAIChatCompletionCreateParams
    ) -> NexusflowAIChatCompletionCreateParams:
        if not create_params.get("tools") and not create_params.get("response_format"):
            return create_params

        create_params = deepcopy(create_params)
        messages = create_params["messages"]
        if create_params.get("response_format"):
            response_format_message = f"""Your aim is to process the given unstructured input data and return the output based on the instructions and the response_format schema provided. Provide only the raw output data in valid JSON format based on the given response_format. All values for JSON attributes should be on quotes and never give incomplete responses. Remember, your responses MUST be valid parsable JSON and MUST match the schema specified in response_format. Do not give any introduction in the front. Your response should ONLY contain the JSON
Response format: {create_params.get('response_format')}"""
            messages.insert(0, {"role": "system", "content": response_format_message})

        create_params = self.qwen_translator.convert_request_to_qwen_params(
            create_params
        )
        messages = create_params["messages"]

        if messages and messages[0].role != "system":
            messages = [
                Message(
                    role="system", content=[ContentItem(text=DEFAULT_SYSTEM_MESSAGE)]
                )
            ] + messages

        validate_num_fncall_results(messages, support_multimodal_input=False)
        messages = self.fncall_prompt.preprocess_fncall_messages(
            messages=messages,
            functions=create_params.get("tools") or [],
            lang="en",
            parallel_function_calls=create_params.get("parallel_tool_calls", True),
            function_choice=self._openai_tool_choice_to_qwen_function_choice(
                create_params
            ),
        )

        messages = [
            format_as_text_message(msg, add_upload_info=False) for msg in messages
        ]
        messages = simulate_response_completion_with_chat(messages)

        create_params["messages"] = [m.model_dump() for m in messages]

        create_params["stop"] = self.qwen_translator.adjust_stop_words_for_qwen_agent(
            create_params.get("stop", [])
        )

        create_params.pop("tools", None)
        create_params.pop("tool_choice", None)
        create_params.pop("parallel_tool_calls", None)

        return create_params

    @staticmethod
    def _openai_tool_choice_to_qwen_function_choice(
        create_params: NexusflowAIChatCompletionCreateParams,
    ) -> str:
        function_choice = create_params.get("tool_choice", "auto")
        if isinstance(function_choice, dict):
            function_choice = function_choice["function"]["name"]

        return function_choice

    def postprocess(
        self,
        create_params: NexusflowAIChatCompletionCreateParams,
        response: NexusflowAIChatCompletion,
    ) -> NexusflowAIChatCompletion:
        if not create_params.get("tools") and not create_params.get("response_format"):
            return response

        content = response.choices[0].message.content

        if create_params.get("response_format"):
            try:
                content = json.dumps(json_repair.loads(content))
            except JSONDecodeError:
                pass

        response_message = Message(
            role="assistant",
            content=[ContentItem(text=content)],
        )

        processed_messages = _postprocess_stop_words(
            messages=[response_message],
            stop=self.qwen_translator.adjust_stop_words_for_qwen_agent(
                create_params.get("stop", [])
            ),
        )

        messages = self.fncall_prompt.postprocess_fncall_messages(
            messages=processed_messages,
            parallel_function_calls=create_params.get("parallel_tool_calls", True),
            function_choice=self._openai_tool_choice_to_qwen_function_choice(
                create_params
            ),
        )
        chat_completion_message = (
            self.qwen_translator.convert_qwen_messages_to_openai_message(
                messages=messages,
                tools=create_params.get("tools"),
                tool_choice=create_params.get("tool_choice"),
            )
        )

        return NexusflowAIChatCompletion(
            id=response.id,
            choices=[
                NexusflowAIChatCompletionChoice(
                    finish_reason="tool_calls",
                    index=0,
                    message=chat_completion_message,
                )
            ],
            created=response.created,
            model=response.model,
            object="chat.completion",
            system_fingerprint=None,
            usage=response.usage,
            raw_prompt=response.raw_prompt,
        )



from typing import Any, Dict, List, Tuple, Optional, Union

from dataclasses import dataclass

from copy import deepcopy

from keyword import iskeyword

from ast import AST, parse as ast_parse
from ast import Call, Import, ImportFrom, NodeVisitor, Name

import json
import json_repair

from openai.types.chat import (
    ChatCompletionToolParam,
    ChatCompletionMessageToolCallParam,
    CompletionCreateParams,
    ChatCompletionMessageParam,
    ChatCompletionContentPartParam,
    ChatCompletionContentPartImageParam,
    ChatCompletionContentPartInputAudioParam,
    ChatCompletionContentPartRefusalParam,
)
from openai.types.chat.chat_completion_tool_choice_option_param import (
    ChatCompletionNamedToolChoiceParam,
)

from nexusflowai.types.chat_completion_message import NexusflowAIChatCompletionMessage


ALLOWED_ROLES = [
    "system",
    "tool",
    "user",
    "assistant",
]

ALLOWED_IMPORTS = [
    "typing",
    "itertools",
    "collections",
    "tabulate",
    "dataclasses",
    "requests",
    "enum",
    "__future__",
]
DISALLOWED_FUNCTIONS = ["eval", "exec", "setattr", "locals", "globals"]


class SecurityVisitor(NodeVisitor):
    def visit_Call(self, node: Call):
        if isinstance(node.func, Name):
            func_name = node.func.id
            if func_name in DISALLOWED_FUNCTIONS:
                raise ValueError(
                    f"Found dangerous call to {func_name} at line {node.lineno}. Disallowed functions: {DISALLOWED_FUNCTIONS}"
                )

        self.generic_visit(node)

    def visit_Import(self, node: Import) -> Any:
        import_names = [a.name for a in node.names]
        self.check_imports(node, import_names)

        self.generic_visit(node)

    def visit_ImportFrom(self, node: ImportFrom) -> Any:
        import_names = [node.module]
        self.check_imports(node, import_names)

        self.generic_visit(node)

    def check_imports(
        self, node: Union[Import, ImportFrom], import_names: List[str]
    ) -> None:
        disallowed_import_names = [n for n in import_names if n not in ALLOWED_IMPORTS]
        if disallowed_import_names:
            raise ImportError(
                f"Not allowed to import {disallowed_import_names} at line {node.lineno}. Allowed imports include {ALLOWED_IMPORTS}"
            )


class CodeExecutionHelper:
    def clean_input(self, source: str) -> Tuple[str, Optional[AST]]:
        source = source.strip()
        if len(source) == 0:
            return source, None

        try:
            tree = ast_parse(source)
        except SyntaxError as e:
            new_source = self._try_fix_syntax_error(e, source)
            try:
                tree = ast_parse(new_source)
                source = new_source
            except:  # pylint: disable=bare-except
                tree = None
        except:  # pylint: disable=bare-except
            tree = None

        return source, tree

    @staticmethod
    def _try_fix_syntax_error(error: SyntaxError, source: str) -> str | None:
        res = None
        if not error.args:
            return res

        error_message = error.args[0]
        res = source
        if error_message == "'(' was never closed":
            res = f"{source})"
        elif error_message == "unmatched ')'":
            res = source.removesuffix(")")
        else:
            """
            Unescaped quotes can manifest as a variety of error messages including:
            - unterminated string literal
            - Perhaps you forgot a comma?

            Hence we use this as a fallback case
            """

            possible_first_quotes = {'="': '")', "='": "')"}
            for possible_first_quote, last_quote in possible_first_quotes.items():
                if possible_first_quote not in source:
                    continue

                res = source.split(possible_first_quote, maxsplit=1)
                res = "='''".join(res)
                res = res.split(last_quote, maxsplit=1)
                res = "''')".join(res)

        return res

    def check_security(self, tree: AST) -> None:
        security_visitor = SecurityVisitor()
        try:
            security_visitor.visit(tree)
        except ValueError as e:
            raise TypeError from e


"""
A smattering of symbols that are invalid for function and argument names
These are all replaced with "_" downstream.
This is a translation table used in `str.translate`.
"""
INVALID_NAME_CHARS = str.maketrans(dict.fromkeys(" !%&(*+,-/<=>@[^|~", "_"))


def clean_str(s: str) -> str:
    s = s.translate(INVALID_NAME_CHARS)
    if iskeyword(s):
        s = f"_{s}"
    return s


def clean_raw_response(raw_response: str) -> str:
    response = raw_response.split("Thought:")[0]
    response = response.strip()
    return response


@dataclass
class CleanedFunction:
    dirty_name: str
    dirty_argument_names: List[str]

    def __post_init__(self) -> None:
        self.clean_name = clean_str(self.dirty_name)
        self.clean_argument_names = list(map(clean_str, self.dirty_argument_names))

        self.argument_names_d2c = dict(
            zip(self.dirty_argument_names, self.clean_argument_names)
        )
        self.argument_names_c2d = dict(
            zip(self.clean_argument_names, self.dirty_argument_names)
        )

    def clean_tool_param(
        self, tool: ChatCompletionToolParam
    ) -> ChatCompletionToolParam:
        tool = deepcopy(tool)

        function = tool["function"]
        function["name"] = self.clean_name

        properties = function["parameters"].get("properties")
        if properties:
            properties = {
                self.argument_names_d2c.get(k, k): v for k, v in properties.items()
            }
            function["parameters"]["properties"] = properties

        required = function["parameters"].get("required")
        if required:
            function["parameters"]["required"] = [
                self.argument_names_d2c.get(k, k) for k in required
            ]

        return tool

    def clean_tool_call_param(
        self, tool_call: ChatCompletionMessageToolCallParam
    ) -> ChatCompletionMessageToolCallParam:
        tool_call = deepcopy(tool_call)

        function = tool_call["function"]
        function["name"] = self.clean_name

        arguments = json_repair.loads(function["arguments"])
        arguments = {self.argument_names_d2c.get(k, k): v for k, v in arguments.items()}
        function["arguments"] = json.dumps(arguments)

        return tool_call

    def dirty_function_call_dict(
        self, clean_name: str, clean_kwargs: Dict[str, Any]
    ) -> Tuple[str, Dict[str, Any]]:
        assert self.clean_name == clean_name
        return self.dirty_name, {
            self.argument_names_c2d.get(k, k): v for k, v in clean_kwargs.items()
        }

    @classmethod
    def from_tool_param(cls, tool: ChatCompletionToolParam) -> "CleanedFunction":
        function = tool["function"]
        function_properties = function["parameters"].get("properties", dict())
        cf = cls(
            dirty_name=function["name"],
            dirty_argument_names=list(function_properties),
        )

        return cf


def resolve_tools_and_tool_choice(
    tools: List[ChatCompletionToolParam],
    tool_choice: ChatCompletionNamedToolChoiceParam,
) -> Tuple[List[ChatCompletionToolParam], ChatCompletionNamedToolChoiceParam]:
    """
    Supported tool choices:
    - `required`: [Not with OpenAI parity] The model will always output a function call except when it thinks the trajectory is finished (this is the way the model is trained).
    - `auto`: Same as `required`.
    - `none`: The model will see all of the tools but is forced to chat.
    - `{"type": "function": "function": {"name": <>}}`: Force a function call
    - `required_with_chat`: The model will see all of the tools and an additional chat function. Including the additional chat function, this tool choice then has the same behavior as `required`
    """
    if not tool_choice:
        return tools, tool_choice

    tools = tools.copy()
    if tool_choice in ["required_with_chat", "none"]:
        if tool_choice == "required_with_chat":
            tool_choice = "required"
        elif tool_choice == "none":
            tool_choice = {"type": "function", "function": {"name": "chat"}}
        tools.append(
            {
                "type": "function",
                "function": {
                    "name": "chat",
                    "description": """This function is used to chat with the user. Use this function to communicate the required information to the user, ask the user for more information, or to let the user know that something has been accomplished.

Don't make assumptions about what values to plug into functions. Use this function to ask for clarification if a user request is ambiguous.

You must call at least one function. If you don't know which function to call, please call this chat function. You can use this function as an argument in nested calls to represent information that is necessary.""",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "message": {
                                "type": "string",
                                "description": "A clear message that the user will read.",
                            },
                        },
                        "required": ["message"],
                        "additionalProperties": False,
                    },
                },
            }
        )

    return tools, tool_choice


def get_tool_choice_prefix(
    tool_choice: ChatCompletionNamedToolChoiceParam | None,
    dirty_function_name_to_cf: Dict[str, CleanedFunction],
) -> str | None:
    if not isinstance(tool_choice, dict):
        return None

    tool_choice: ChatCompletionNamedToolChoiceParam
    dirty_function_name = tool_choice["function"]["name"]

    assert (
        dirty_function_name in dirty_function_name_to_cf
    ), f"Did not find any tools with the `tool_choice` function name `{dirty_function_name}`"

    cleaned_name = dirty_function_name_to_cf[dirty_function_name].clean_name

    return cleaned_name


def _parse_content_parts(
    content_parts: List[
        Union[ChatCompletionContentPartParam, ChatCompletionContentPartRefusalParam]
    ]
) -> List[str]:
    if content_parts == []:
        return []

    content_strs = []

    for part in content_parts:
        part_type = part["type"]

        match part_type:
            case "text":
                text = part["text"]
                content_strs.append(text)

            case "image_url":
                raise NotImplementedError(
                    f"{ChatCompletionContentPartImageParam} is not supported!"
                )

            case "input_audio":
                raise NotImplementedError(
                    f"{ChatCompletionContentPartInputAudioParam} is not supported!"
                )

            case "refusal":
                raise NotImplementedError(
                    f"{ChatCompletionContentPartRefusalParam} ist not supported!"
                )

    return content_strs


def _flatten_chat_completion_create_params_messages_content_parts(
    create_params: CompletionCreateParams,
) -> CompletionCreateParams:
    """
    Expands ContentPartParams to individual messages.
    Used for chat flows in backend.
    """

    def _parse_chat_message_content_parts(
        message: ChatCompletionMessageParam,
        content_parts: List[
            Union[ChatCompletionContentPartParam, ChatCompletionContentPartRefusalParam]
        ],
    ) -> List[ChatCompletionMessageParam]:
        # Expanded messages:
        # The last message is original message iwth all calls / inputs.
        # The other messages only have the role and content information.
        messages: List[ChatCompletionMessageParam] = [
            *[
                {
                    "content": None,
                    "role": message["role"],
                }
                for _ in range(len(content_parts) - 1)
            ],
            message,
        ]
        if "name" in message:
            for sub_message in messages:
                sub_message["name"] = message["name"]

        for i, content_str in enumerate(_parse_content_parts(content_parts)):
            messages[i]["content"] = content_str

        return messages

    def _expand_chat_message_content(
        message: ChatCompletionMessageParam,
    ) -> List[ChatCompletionMessageParam]:
        content = message.get("content")

        if isinstance(content, str):
            return [message]

        if content is None:
            content = []

        result = _parse_chat_message_content_parts(message, content)

        return result

    def _expand_chat_messages(
        messages: List[ChatCompletionMessageParam],
    ) -> List[ChatCompletionMessageParam]:
        conversation = []

        for message in messages:
            sub_messages = _expand_chat_message_content(message)
            conversation.extend(sub_messages)

        return conversation

    create_params["messages"] = _expand_chat_messages(create_params["messages"])

    return create_params


# Tool call ID to nexusflowai_extras
tci_to_nexusflowai_extras: Dict[str, Dict[str, Any]] = dict()


def add_nexusflowai_extras_to_messages(
    messages: List[ChatCompletionMessageParam],
) -> List[ChatCompletionMessageParam]:
    messages = deepcopy(messages)
    for message in messages:
        if message["role"] != "assistant" or not message.get("tool_calls"):
            continue

        for tool_call in message["tool_calls"]:
            if tci_to_nexusflowai_extras.get(tool_call["id"]):
                message["nexusflowai_extras"] = tci_to_nexusflowai_extras[
                    tool_call["id"]
                ]
                break

    return messages


def store_nexusflowai_extras_in_memory(
    message: NexusflowAIChatCompletionMessage,
) -> None:
    tool_calls = message.tool_calls or []
    for tool_call in tool_calls:
        tci_to_nexusflowai_extras[tool_call.id] = message.nexusflowai_extras
